"""Principal nested spheres analysis."""

from sklearn.base import BaseEstimator, TransformerMixin

from .pns import from_unit_sphere, pns, proj, to_unit_sphere

__all__ = [
    "PNS",
]


class PNS(TransformerMixin, BaseEstimator):
    """Principal nested spheres (PNS) analysis [1]_.

    Parameters
    ----------
    n_components : int, default=2
        Number of components to keep.
        Data are transformed onto unit hypersphere embedded in this dimension.
    tol : float, default=1e-3
        Optimization tolerance.

    Attributes
    ----------
    embedding_ : ndarray of shape (n_samples, n_components)
        Stores the embedding vectors.
    v_ : list of (n_features - 1) arrays
        Principal directions of nested spheres.
    r_ : ndarray of shape (n_features - 1,)
        Principal radii of nested spheres.

    References
    ----------
    .. [1] Jung, Sungkyu, Ian L. Dryden, and James Stephen Marron.
       "Analysis of principal nested spheres." Biometrika 99.3 (2012): 551-568.

    Examples
    --------
    >>> from skpns import PNS
    >>> from skpns.util import circular_data
    >>> X = PNS(n_components=2).fit_transform(circular_data())
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... plt.plot(*X.T, "x")
    ... plt.gca().set_aspect("equal")
    """

    def __init__(self, n_components=2, tol=1e-3):
        self.n_components = n_components
        self.tol = tol

    def _fit_transform(self, X):
        self.v_ = []
        self.r_ = []

        D = X.shape[1]
        pns_ = pns(X, self.tol)
        for _ in range(D - self.n_components):
            v, r, A = next(pns_)
            self.v_.append(v)
            self.r_.append(r)
        self.embedding_ = A

    def fit(self, X, y=None):
        """Find principal nested spheres for data X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data on (n_features - 1)-dimensional hypersphere.
        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : object
            Returns a fitted instance of self.
        """
        self._fit_transform(X)
        return self

    def fit_transform(self, X, y=None):
        """Fit the model from data in X and transform X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data on (n_features - 1)-dimensional hypersphere.
        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            X transformed in the new space.
        """
        self._fit_transform(X)
        return self.embedding_

    def transform(self, X):
        """Transform X onto the fitted subsphere.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data on (n_features - 1)-dimensional hypersphere.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
            X transformed in the new space.
        """
        if X.shape[1] != len(self.v_[0]):
            raise ValueError("Input dimension does not match the fitted dimension.")

        for v, r in zip(self.v_, self.r_):
            A = proj(X, v, r)
            X = to_unit_sphere(A, v, r)
        return X

    def to_hypersphere(self, X):
        """Transform the low-dimensional data into the original hypersphere.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_components)

        Returns
        -------
        X_new : array-like of shape (n_samples, n_features)
        """
        for v, r in zip(reversed(self.v_), reversed(self.r_)):
            X = from_unit_sphere(X, v, r)
        return X



"""
Utilities
---------
"""

import numpy as np

__all__ = [
    "circular_data",
    "unit_sphere",
    "circle",
]


def circular_data():
    """Circular data in 3D.

    Returns
    -------
    ndarray of shape (100, 3)
        Data coordinates.
    """
    t = np.random.uniform(0.1 * np.pi, 0.2 * np.pi, 100)
    p = np.random.uniform(0, 3 * np.pi / 2, 100)
    circle = np.array([np.sin(t) * np.cos(p), np.sin(t) * np.sin(p), np.cos(t)]).T

    v = np.array([1 / np.sqrt(3), -1 / np.sqrt(3), 1 / np.sqrt(3)])
    north_pole = np.array([0.0, 0.0, 1.0])
    u = v - north_pole
    u /= np.linalg.norm(u)
    H = np.eye(3) - 2 * np.outer(u, u)
    return (H @ circle.T).T


def unit_sphere():
    """Helper function to plot a unit sphere.

    Returns
    -------
    x, y, z : array
        Coordinates for unit sphere.
    """
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)
    x = np.outer(np.cos(u), np.sin(v))
    y = np.outer(np.sin(u), np.sin(v))
    z = np.outer(np.ones_like(u), np.cos(v))
    return x, y, z


def circle(v, theta, n=100):
    """Helper function to plot a circle in 3D.

    Parameters
    ----------
    v : (3,) array
        Unit vector to center of circle in 3D.
    theta : scalar
        Geodesic distance.
    n : int, default=100
        Number of points.
    """
    phi = np.linspace(0, 2 * np.pi, n)
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.full_like(phi, np.cos(theta))
    circle = np.stack([x, y, z], axis=1)

    north_pole = np.array([0.0, 0.0, 1.0])
    u = v - north_pole
    u /= np.linalg.norm(u)
    H = np.eye(3) - 2 * np.outer(u, u)
    return H @ circle.T



"""
PNS
---

Functions to find principal nested spheres.
"""

import numpy as np
from scipy.optimize import least_squares

__all__ = [
    "pns",
    "pss",
    "proj",
    "residual",
    "to_unit_sphere",
    "from_unit_sphere",
    "Exp",
    "Log",
]


def pns(x, tol=1e-3):
    """Principal nested spheres analysis.

    Parameters
    ----------
    x : (N, d+1) real array
        Data on d-sphere.
    tol : float, default=1e-3
        Convergence tolerance in radian.

    Yields
    ------
    v : 1-D ndarray
        Principal axis.
    r : scalar
        Principal geodesic distance.
    x : (N, d-i) real array
        Data transformed onto low-dimensional unit hypersphere.

    Examples
    --------
    >>> from skpns.pns import pns, from_unit_sphere
    >>> from skpns.util import circular_data, unit_sphere, circle
    >>> x = circular_data()
    >>> v, r, A = next(pns(x))
    >>> import matplotlib.pyplot as plt  # doctest: +SKIP
    ... ax = plt.figure().add_subplot(projection='3d', computed_zorder=False)
    ... ax.plot_surface(*unit_sphere(), color='skyblue', alpha=0.6, edgecolor='gray')
    ... ax.scatter(*x.T, marker=".")
    ... ax.scatter(*from_unit_sphere(A, v, r).T, marker="x")
    ... ax.plot(*circle(v, r), color="tab:red")
    """
    d = x.shape[1] - 1

    for _ in range(1, d):
        v, r = pss(x, tol)
        A = proj(x, v, r)
        x = to_unit_sphere(A, v, r)
        yield v, r, x

    v, r = pss(x, tol)
    x = np.full((len(x), 1), 0)
    yield v, r, x


def pss(x, tol=1e-3):
    """Find the principal subsphere, i.e., lower the dimension by one.

    Parameters
    ----------
    x : (N, d+1) real array
        Data on d-sphere.
    tol : float, default=1e-3
        Convergence tolerance in radian.

    Returns
    -------
    v : (d+1,) real array
        Principal axis.
    r : scalar
        Principal geodesic distance.
    """
    _, D = x.shape
    if D <= 1:
        raise ValueError("Data must be on at least 1-sphere.")
    elif D == 2:
        r = 0
        v = np.mean(x, axis=0)
        v /= np.linalg.norm(v)
    else:
        pole = np.array([0] * (D - 1) + [1])
        R = np.eye(D)
        _x = x
        v, r = _pss(_x)
        while np.arccos(np.dot(pole, v)) > tol:
            # Rotate so that v becomes the pole
            _x, _R = _rotate(_x, v)
            v, r = _pss(_x)
            R = R @ _R.T
        v = R @ v  # re-rotate back
    return v, r


def proj(x, v, r):
    """Minimum-geodesic projection of points to subsphere.

    Parameters
    ----------
    x : (N, d+1) real array
        Data on unit d-sphere.
    v : (d+1,) real array
        Subsphere axis.
    r : scalar
        Subsphere geodesic distance.

    Returns
    -------
    A : (N, d+1) real array
        Points projected onto the small-subsphere on unit d-sphere.
    """
    geod = np.arccos(x @ v)[..., np.newaxis]
    return (np.sin(r) * x + np.sin(geod - r) * v) / np.sin(geod)


def residual(x, v, r):
    """Signed residuals of dimension reduction to subsphere.

    Parameters
    ----------
    x : (N, d+1) real array
        Data on d-sphere.
    v : (d+1,) real array
        Subsphere axis.
    r : scalar
        Subsphere geodesic distance.

    Returns
    -------
    xi : (N,) array
        Signed residual.
    """
    _, D = x.shape
    if D <= 1:
        raise ValueError("Data must be on at least 1-sphere.")
    elif D == 2:
        xi = np.arctan2(x @ (v @ [[0, 1], [-1, 0]]), x @ v)
    else:
        xi = np.arccos(np.dot(x, v.T)) - r
    return xi


def _R(v):
    a = np.zeros_like(v)
    a[-1] = 1.0
    b = v
    c = b - a * (a @ b)
    c /= np.linalg.norm(c)

    A = np.outer(a, c) - np.outer(c, a)
    theta = np.arccos(v[-1])
    Id = np.eye(len(A))
    R = Id + np.sin(theta) * A + (np.cos(theta) - 1) * (np.outer(a, a) + np.outer(c, c))
    return R


def to_unit_sphere(x, v, r):
    """Transform projected data on subsphere to unit hypersphere.

    Parameters
    ----------
    x : (N, d+1) real array
        Data projected on subsphere.
    v : (d+1,) real array
        Subsphere axis.
    r : scalar
        Subsphere geodesic distance.

    Returns
    -------
    (N, d) real array
        Data points on unit hypersphere.
    """
    R = _R(v)
    return (1 / np.sin(r) * R[:-1:, :] @ x.T).T


def from_unit_sphere(x, v, r):
    """Transform data on unit hypersphere to projected data on subsphere.

    Parameters
    ----------
    x : (N, d) real array
        Data points on unit hypersphere.
    v : (d+1,) real array
        Subsphere axis.
    r : scalar
        Subsphere geodesic distance.

    Returns
    -------
    (N, d+1) real array
        Data projected on subsphere.
    """
    R = _R(v)
    vec = np.hstack([np.sin(r) * x, np.full(len(x), np.cos(r)).reshape(-1, 1)])
    return (R.T @ vec.T).T


def _pss(pts):
    # Projection
    x_dag = Log(pts)
    v_dag_init = np.mean(x_dag, axis=0)
    r_init = np.mean(np.linalg.norm(x_dag - v_dag_init, axis=1))
    init = np.concatenate([v_dag_init, [r_init]])
    # Optimization
    opt = least_squares(_loss, init, args=(x_dag,), method="lm").x
    v_dag_opt, r_opt = opt[:-1], opt[-1]
    v_opt = Exp(v_dag_opt.reshape(1, -1)).reshape(-1)
    r_opt = np.mod(r_opt, np.pi)
    return v_opt, r_opt


def _loss(params, x_dag):
    v_dag, r = params[:-1], params[-1]
    return np.linalg.norm(x_dag - v_dag.reshape(1, -1), axis=1) - r


def _rotate(pts, v):
    R = _R(v)
    return (R @ pts.T).T, R


def Exp(z):
    """Exponential map of hypersphere at (0, 0, ..., 0, 1).

    Parameters
    ----------
    z : (N, d) real array
        Vectors on tangent space.

    Returns
    -------
    (N, d+1) real array
        Points on d-sphere.
    """
    norm = np.linalg.norm(z, axis=1)[..., np.newaxis]
    return np.hstack([np.sin(norm) / norm * z, np.cos(norm)])


def Log(x):
    """Log map of hypersphere at (0, 0, ..., 0, 1).

    Parameters
    ----------
    x : (N, d+1) real array
        Points on d-sphere.

    Returns
    -------
    (N, d) real array
        Vectors on tangent space.
    """
    thetas = np.arccos(x[:, -1:])
    return thetas / np.sin(thetas) * x[:, :-1]



import numpy as np

from skpns import PNS
from skpns.util import circular_data

np.random.seed(0)


def test_PNS_transform():
    X = circular_data()
    pns = PNS(n_components=2)
    Xnew_1 = pns.fit_transform(X)
    Xnew_2 = pns.transform(X)
    assert np.all(Xnew_1 == Xnew_2)
